"""Utilities."""

import logging
import time
from collections.abc import Callable
from functools import partial
from typing import Any, TypeVar

R = TypeVar("R")


def timing(func: Callable[..., R] | None = None, level: int = logging.INFO) -> Any:
    """Log the elapsed time of a function.

    Decorator can be used with or without arguments. Eg: `@timing` or `@timing(level=logging.DEBUG)`.

    Parameters
    ----------
    func
        The wrapped function to log the elapsed time of.
    level
        Log level to user. Default: INFO.

    Returns
    -------
    Any
        The function wrapped to log timing.
    """
    if func is None:
        return partial(timing, level=level)

    def wrapper(
        *args: list[Any],
        **kwargs: dict[str, Any],
    ) -> Any:
        """Log elapsed time of the wrapped function."""
        start = time.perf_counter()
        result = func(*args, **kwargs)
        end = time.perf_counter()
        logging.getLogger(__name__).log(level, f"Timing for '{func.__name__}': {end - start:.3f}s")
        return result

    return wrapper



"""Datetime utilities."""

import datetime


def utcnow() -> datetime.datetime:
    """Generate now as a datetime with UTC time zone info.

    The `datetime.datetime.utcnow()` function from the standard library does not include time zone information by
    default. https://docs.python.org/3/library/datetime.html#datetime.datetime.now
    """
    return datetime.datetime.now(tz=datetime.UTC)



"""Array utilities."""

from typing import TypeVar

T = TypeVar("T")


def chunkify(seq: list[T], s: int) -> list[list[T]]:
    """Break a list into chunks of size S."""
    if s < 1:
        raise ValueError(f"Size must be > 0, invalid value {s}")
    return [seq[pos : pos + s] for pos in range(0, len(seq), s)]



"""String utilities.

More info on casing naming conventions:
https://en.wikipedia.org/wiki/Naming_convention_(programming)
"""

import re
import unicodedata


def _normalize(string: str, unicode: bool = True, compact_spaces: bool = True) -> str:
    """Normalize a string for casing.

    Replace non-alphanumeric characters (except whitespace) with spaces. Leading and trailing whitespace will be stripped.

    - If 'unicode' is True, unicode characters will be preserved.
    - If 'unicode' is False unicode characters will be by ASCII characters.
    - If unicode characters can't be replaced the will be replaced by a space.
    - If 'compact_spaces' is True, multiple consecutive spaces will be reduced to a single space.

    Parameters
    ----------
    string
        The input string to be normalized.
    unicode
        If True, allows Unicode characters in the output string. If False, only ASCII characters are allowed (default is True).
    compact_spaces
        If True, multiple consecutive spaces are reduced to a single space (default is True).

    Returns
    -------
    str
        Returns a transformed string.
    """
    if unicode:
        # Normalize the string to Normalization Form Compatibility Composition (NFKC).
        # This will replace multiple representation by a normalized one. E.g. 'ö' can have two representations.
        value = unicodedata.normalize("NFKC", string)
        text = re.sub(r"[^\w\s]|_", " ", value, flags=re.UNICODE)
    else:
        # Normalize the string to Normalization Form Compatibility Decomposition (NFKD).
        # This will replace the diacritics by ASCII characters. E.g. 'ö' will be replaced by 'o' and 'ì' by 'i'.
        value = unicodedata.normalize("NFKD", string).encode("ascii", "ignore").decode("ascii")
        text = re.sub(r"[^\w\s]|_", " ", value)

    # Replace multiple spaces with a single space
    return re.sub(r"\s+", " ", text).strip() if compact_spaces else text.strip()


def kebab_case(string: str, scream: bool = False, unicode: bool = True, compact_spaces: bool = True) -> str:
    """Convert a string to kebab-case.

    Parameters
    ----------
    string
        Input string to transform.
    scream
        Convert the output to uppercase.
    unicode
        If True, allows Unicode characters in the output string. If False, only ASCII characters are allowed (default is True).
    compact_spaces
        If True, multiple consecutive spaces are reduced to a single space (default is True).

    Returns
    -------
    str
        Returns a transformed string.
    """
    text = _normalize(string, unicode=unicode, compact_spaces=compact_spaces).replace(" ", "-")
    return text.lower() if not scream else text.upper()


def slugify(string: str) -> str:
    """Create a slug from a given string.

    Normalize strings to a 'slug'. Can be used to format URL's or resource names (eg: Database name). If you want more
     flexibility, you can use 'kebab_case'.

    Parameters
    ----------
    string
        Input string to transform.

    Returns
    -------
    str
        Returns a transformed string.
    """
    return kebab_case(string, unicode=False, compact_spaces=True)


def _dromedary_case(string: str, upper: bool = False, join_char: str = "", unicode: bool = True) -> str:
    """Convert a string to dromedaryCase/DromedaryCase.

    Can be used for camelCase, UpperCamelCase, and PascalCase. PascalCase and UpperCamelCase are
    interchangeable. PascalCase originates from Pascal programming language, which popularized this
    style.
    """
    words = _normalize(string, unicode=unicode, compact_spaces=True).split(" ")
    camel_case_words = [words[0].lower() if not upper else words[0].capitalize()] + [
        word.capitalize() for word in words[1:]
    ]
    return join_char.join(camel_case_words)


def camel_case(string: str, unicode: bool = True) -> str:
    """Convert a string to camelCase.

    Parameters
    ----------
    string
        Input string to transform.
    unicode
        If True, allows Unicode characters in the output string. If False, only ASCII characters are allowed (default is True).

    Returns
    -------
    str
        Returns a transformed string.
    """
    return _dromedary_case(string, upper=False, unicode=unicode)


def pascal_case(string: str, unicode: bool = True) -> str:
    """Convert a string to PascalCase.

    Parameters
    ----------
    string
        Input string to transform.
    unicode
        If True, allows Unicode characters in the output string. If False, only ASCII characters are allowed (default is True).

    Returns
    -------
    str
        Returns a transformed string.
    """
    return _dromedary_case(string, upper=True, unicode=unicode)


def train_case(string: str, unicode: bool = False) -> str:
    """Convert a string to train-case.

    Also known as HTTP-Header-Case, this style is used for HTTP headers. For HTTP headers it is recommended to use
    ASCII characters since all parties need to be aligned (sender, receiver, and intermediaries). Hence, 'unicode' is
    False by default.

    Parameters
    ----------
    string
        Input string to transform.
    unicode
        If True, allows Unicode characters in the output string. If False, only ASCII characters are allowed (default is False).

    Returns
    -------
    str
        Returns a transformed string.
    """
    return _dromedary_case(string, upper=True, join_char="-", unicode=unicode)


def snake_case(string: str, scream: bool = False, unicode: bool = True, compact_spaces: bool = True) -> str:
    """Convert a string to snake_case.

    Parameters
    ----------
    string
        Input string to transform.
    scream
        Convert the output to uppercase.
    unicode
        If True, allows Unicode characters in the output string. If False, only ASCII characters are allowed (default is True).
    compact_spaces
        If True, multiple consecutive spaces are reduced to a single space (default is True).

    Returns
    -------
    str
        Returns a transformed string.
    """
    text = _normalize(string, unicode=unicode, compact_spaces=compact_spaces).replace(" ", "_")
    return text.lower() if not scream else text.upper()


def truncate(string: str, number: int, /, suffix: str = "...") -> str:
    """Truncate a string to a certain number of characters."""
    if number <= 0:
        raise ValueError("Number must be a positive integer.")
    if len(string) <= number:
        return string
    return f"{string[: number - 1]}{suffix}"



"""Utility functions for working with bytes."""

_BASE_1000: int = 1000
_BASE_1024: int = 1024

# Units decimal: base 1000
_UNITS_DECIMAL_SHORT: list[str] = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
_UNITS_DECIMAL_LONG: list[str] = [
    "Bytes",
    "Kilobytes",
    "Megabytes",
    "Gigabytes",
    "Terabytes",
    "Petabytes",
    "Exabytes",
    "Zettabytes",
    "Yottabytes",
]

# Unit binary: base 1024
_UNITS_BINARY_SHORT: list[str] = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"]
_UNITS_BINARY_LONG: list[str] = [
    "Bytes",
    "Kibibytes",
    "Mebibytes",
    "Gibibytes",
    "Tebibytes",
    "Pebibytes",
    "Exbibytes",
    "Zebibytes",
    "Yobibytes",
]

# decimal-short, decimal-long, binary-short, binary-long
_FORMATS: set[str] = {"ds", "dl", "bs", "bl"}


def pretty_bytes(size: int, fmt: str = "ds", /, precision: int = 2) -> str:
    """Convert a size in bytes to a human-readable string.

    Available formats:
    - ds: decimal-short (e.g., 1.23 KB)
    - dl: decimal-long (e.g., 1.23 Kilobytes)
    - bs: binary-short (e.g., 1.23 KiB)
    - bl: binary-long (e.g., 1.23 Kibibytes)

    Parameters
    ----------
    size : int
        The size in bytes.
    fmt : str
        The format to use. One of "ds", "dl", "bs", "bl".
    precision : int
        The number of decimal places to round to.

    Returns
    -------
    str
        The formatted disk size as a human-readable string.
    """
    if size < 0:
        raise ValueError("Size must be a non-negative integer.")
    if fmt not in _FORMATS:
        raise ValueError(
            f"Format must be one of {_FORMATS}.\n  ds: decimal short (e.g. MB)\n  dl: decimal long (e.g. Megabytes)\n  bs: binary short (e.g. MiB)\n  bl: binary long (e.g. Mebibytes) "
        )
    base = _BASE_1000 if fmt in {"ds", "dl"} else _BASE_1024
    units = {
        "ds": _UNITS_DECIMAL_SHORT,
        "dl": _UNITS_DECIMAL_LONG,
        "bs": _UNITS_BINARY_SHORT,
        "bl": _UNITS_BINARY_LONG,
    }.get(fmt, ["N/A"])
    index = 0
    size_: float = float(size)
    while size_ >= base and index < len(units) - 1:
        size_ /= base
        index += 1
    return f"{size_:.{precision}f} {units[index]}"



"""Orval package."""

from orval.arrays import chunkify
from orval.byte_utils import pretty_bytes
from orval.datetimes import utcnow
from orval.strings import camel_case, kebab_case, pascal_case, slugify, snake_case, train_case, truncate
from orval.utils import timing

__all__ = [
    "camel_case",
    "chunkify",
    "kebab_case",
    "pascal_case",
    "pretty_bytes",
    "slugify",
    "snake_case",
    "timing",
    "train_case",
    "truncate",
    "utcnow",
]
